"""Telemetry configuration module."""

from thoughtful.supervisor.telemetry.config.builder import (
    TelemetryConfig,
    TelemetryConfigBuilder,
)
from thoughtful.supervisor.telemetry.config.context import (
    TelemetryContext,
    clear_global_telemetry_context,
    get_global_telemetry_context,
    get_service_name,
    mark_telemetry_initialized_early,
    reset_telemetry_early_flag,
    set_global_telemetry_context,
    set_service_name,
    was_telemetry_initialized_early,
)

__all__ = [
    "TelemetryConfig",
    "TelemetryConfigBuilder",
    "TelemetryContext",
    "get_global_telemetry_context",
    "set_global_telemetry_context",
    "clear_global_telemetry_context",
    "get_service_name",
    "set_service_name",
    "was_telemetry_initialized_early",
    "mark_telemetry_initialized_early",
    "reset_telemetry_early_flag",
]
